/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wF1El.h"
using namespace tbox;
using namespace flow;

F1ElVi::F1ElVi(int _nDim, double alpha, double beta) : Fct1(), nDim(_nDim)
{
    // Sanity check
    if (nDim != 2 && nDim != 3)
    {
        std::stringstream err;
        err << "F1ElVi::nDim should be equal to 2 or 3 but " << nDim << " was given!\n";
        throw std::runtime_error(err.str());
    }
    if (nDim != 3 && beta != 0)
        throw std::runtime_error("F1ElVi::beta can be nonzero only for 3D cases!\n");
    // Set the velocity
    this->update(alpha, beta);
}
/**
 * @brief Update the velocity vector
 * @authors Adrien Crovato
 */
void F1ElVi::update(double alpha, double beta)
{
    if (nDim == 2)
    {
        v << cos(alpha), sin(alpha), 0;
        dv << -sin(alpha), cos(alpha), 0;
    }
    else
    {
        v << cos(alpha) * cos(beta), sin(beta), sin(alpha) * cos(beta);
        dv << -sin(alpha) * cos(beta), 0, cos(alpha) * cos(beta);
    }
}
/**
 * @brief Return the freestream velocity
 * @authors Adrien Crovato
 */
Eigen::Vector3d F1ElVi::eval(Element const &e, std::vector<double> const &u, size_t k) const
{
    return v;
}
/**
 * @brief Return the gradient of freestream velocity with respect to angle of attack
 */
Eigen::Vector3d F1ElVi::evalD(Element const &e, std::vector<double> const &u, size_t k) const
{
    return dv;
}
void F1ElVi::write(std::ostream &out) const
{
    out << "F1ElVi";
}