#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from katoptron.eigenvalues import *
from katoptron.readers import *
import os
import fwk
"""@package docstring
This file is used to investigate the convergence of the SOR (Gauss-Seidel with damping factor in [0,2]).
"""

Verbose = False

algo_inverse = 4


def save_data(name, omegas, rhos):
    tmp = np.zeros((len(omegas), 2))
    tmp[:, 0] = omegas
    tmp[:, 1] = rhos
    np.savetxt(name, tmp)


def Inv_triangular_matrix_1(A):
    """
    This function computes the inverse of a triangular matrix based on
    https://math.stackexchange.com/questions/1003801/inverse-of-an-invertible-upper-triangular-matrix-of-order-3

    This one is slow but precise!
    """
    from scipy.sparse import eye
    n = A.shape[0]
    invA = eye(n)
    tmp = A
    for i in range(1, n - 1):
        invA = invA + tmp
        tmp = tmp * A
    return invA


def Inv_triangular_matrix_2(A):
    """
    This function computes the inverse of a triangular matrix based on
    forward substitution

    This one is fast but not precise!
    """
    from scipy.sparse import eye
    from scipy.sparse import csr_matrix
    n = A.shape[0]
    invA = np.zeros((n, n))
    tmpA = A.todense()
    for i in range(0, n):
        invA[i, :] = invA[i, :] - A[i, 0:(i - 1)].dot(invA[0:(i - 1), :])
        invA[i, i] = 1.
    return csr_matrix(invA)


def Inv_triangular_matrix_3(A, B):
    from scipy.sparse.linalg import spsolve_triangular
    n = A.shape[0]
    X = np.zeros((n, n))

    for i in range(1, n):
        #X[:,[i]] = spsolve_triangular(A, B[:,i])
        tmp = A.tocsr()[0:i, :].tocsc()[:, 0:i]
        X[0:i, [i]] = spsolve_triangular(tmp, B[0:i, i])
    return X


def Inv_triangular_matrix_4(A):
    """
    This function computes the inverse of a triangular matrix based on
    a recursive approach

    This one is fast and precise!
    """
    from scipy.sparse import csr_matrix
    from scipy.sparse import eye
    n = A.shape[0]

    n_max = 100

    if n < n_max:
        return Inv_triangular_matrix_1(eye(n) - A).todense()
    else:
        n1 = n // 2
        invA = np.zeros((n, n))
        invA[0:n1, 0:n1] = Inv_triangular_matrix_4(A[0:n1, 0:n1])
        invA[n1:n, n1:n] = Inv_triangular_matrix_4(A[n1:n, n1:n])
        tmp = invA[n1:n, n1:n] * A[n1:n, 0:n1]
        invA[n1:n, 0:n1] = -tmp.dot(invA[0:n1, 0:n1])
        return invA


def M_Jac(A):
    """
    This function returns the iterative matrix of the Jacobi approach applied on A
    If this matrix has a spectral radius strictly smaller than 1, the optimal 
    damping parameters of the SOR applied on A can be deduced analytically.
    """
    from scipy.sparse import diags
    from scipy.sparse import eye
    invD = diags(1. / A.diagonal(), 0)
    n = A.shape[0]
    return eye(n) - invD * A


def M_SOR(A, omega):
    """
    This function returns the iterative matrix of the SOR approach applied on A.
    """
    from scipy.sparse import tril
    from scipy.sparse import triu
    from scipy.sparse import eye
    from scipy.sparse import diags
    from scipy.sparse.linalg import inv
    from scipy.sparse.linalg import spsolve_triangular

    E = -tril(A, -1)
    F = -triu(A, 1)
    invD = diags(1. / A.diagonal(), 0)

    L = invD * E
    U = invD * F
    I = eye(A.shape[0])

    if algo_inverse == 1:
        x = Inv_triangular_matrix_1(omega * L) * ((1. - omega) * I + omega * U)
    elif algo_inverse == 2:
        x = Inv_triangular_matrix_2(I - omega * L) * (
            (1. - omega) * I + omega * U)
    elif algo_inverse == 3:
        x = Inv_triangular_matrix_3(I - omega * L,
                                    ((1. - omega) * I + omega * U).todense())
    elif algo_inverse == 4:
        x = Inv_triangular_matrix_4(I - omega * L) * (
            (1. - omega) * I + omega * U)
    return x


def spectral_radius_analysis(A,
                             N,
                             timers,
                             tmp_name,
                             omega_min=0.,
                             omega_max=2.):
    """
    This function computes the spectral radii.
    """
    timers['Compute M_Jac'].start()
    m_Jac = M_Jac(A)
    timers['Compute M_Jac'].stop()
    timers['Compute rho_Jac'].start()
    rho_Jac = spectral_radius(m_Jac)
    timers['Compute rho_Jac'].stop()

    if Verbose:
        if rho_Jac >= 1.:
            print('The Jacobi approach does not converge on this matrix as rho(M_Jac) = ' + str(
                rho_Jac))
        else:
            print('The Jacobi approach does converge on this matrix as rho(M_Jac) = ' + str(
                rho_Jac))

    omegas = np.linspace(omega_min, omega_max, N)
    rho_SOR = np.ones(omegas.shape)

    for i in range(0, N):
        if 0. < omegas[i] and omegas[i] < 2.:
            timers['Compute M_SOR'].start()
            m_SOR = M_SOR(A, omegas[i])
            timers['Compute M_SOR'].stop()
            timers['Compute rho_SOR'].start()
            try:
                rho_SOR[i] = spectral_radius(m_SOR)
            except:
                rho_SOR[i:(N + 1)] = 1.
                timers['Compute M_SOR'].stop()
                break
            timers['Compute M_SOR'].stop()

            if Verbose:
                if rho_SOR[i] >= 1.:
                    print('The SOR approach does not converge on this matrix as rho(M_SOR) = ' + str(
                        rho_SOR[i]) + ' with omega = ' + str(omegas[i]))
                else:
                    print('The SOR approach does converge on this matrix as rho(M_SOR) = ' + str(
                        rho_SOR[i]) + ' with omega = ' + str(omegas[i]))
            print(str(((100. * i) // N)) +
                  ' % of the omegas are evaluated ')
            save_data(tmp_name, omegas, rho_SOR)
            if Verbose:
                print(timers)

    return rho_Jac, omegas, rho_SOR


def main():
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    timers = fwk.Timers()

    timers['Read data'].start()
    input_file_dir = file_dir + '/../../../workspace/katoptron_CMAME_tests_plate/'
    filename_K = input_file_dir + 'A_mm_after_bc.txt'
    filename_GT = input_file_dir + 'B_mm.txt'
    filename_rhs = input_file_dir + 'b_mm_after_bc.txt'

    K = read_mm(filename_K)
    K = get_symmetrical_part(K)
    GT = read_mm(filename_GT)
    G = GT.transpose()
    timers['Read data'].stop()
    timers['Compute and apply gamma'].start()
    norm_K = sparse_matrix_2_norm(K)
    norm_GT = sparse_matrix_2_norm(GT)

    b = read_mm(filename_rhs, is_sparse=False)
    b0 = b[0:K.shape[0]]

    gamma = (norm_K / norm_GT)

    Q = K + gamma * G * GT
    timers['Compute and apply gamma'].stop()

    from numpy import linalg as LA

    n = 20
    gs = Gauss_Seidel(K, n=n, damping=1.)
    xn, norm_res_K = gs.dot_with_norm(b0)
    gs = Gauss_Seidel(K, n=n, damping=0.8)
    xn, norm_res_Q = gs.dot_with_norm(b0)

    n_omegas = 50
    omegas = np.linspace(0., 2., n_omegas)
    res_omegas = np.zeros((n_omegas, n+2))
    res_omegas[:, 0] = omegas
    import matplotlib.pyplot as plt
    plt.figure()
    for i in range(0, n_omegas):
        gs = Gauss_Seidel(K, n=n, damping=omegas[i])
        xn, norm_res = gs.dot_with_norm(b0)
        tmp = (norm_res / norm_res[0])
        res_omegas[i, 1:n+2] = np.reshape(tmp[:], (n+1,))
        plt.plot(tmp)
    plt.grid()

    plt.figure()
    plt.plot(omegas, res_omegas[:, 1])
    plt.plot(omegas, res_omegas[:, 2])
    plt.plot(omegas, res_omegas[:, 3])
    print(res_omegas)
    for i in range(1, n+1):
        j = np.argmin(res_omegas[:, i+1])
        print(omegas[j])
    np.savetxt('res_omegas.txt', res_omegas)

    plt.show()

    print(timers)

    '''
    R = approx_Schur_complement(Q, G, GT)

    N = 5

    timers_R = fwk.Timers()
    timers_Q = fwk.Timers()
    rho_Jac_R, omegas_R, rho_SOR_R = spectral_radius_analysis(R,
                                                              N,
                                                              timers_R,
                                                              'tmp_SOR_R.out',
                                                              omega_min=0.,
                                                              omega_max=1.)
    save_data('SOR_R.out', omegas_R, rho_SOR_R)
    print timers_R
    rho_Jac_Q, omegas_Q, rho_SOR_Q = spectral_radius_analysis(Q,
                                                              N,
                                                              timers_Q,
                                                              'tmp_SOR_Q.out',
                                                              omega_min=0.8,
                                                              omega_max=2.)
    save_data('SOR_Q.out', omegas_Q, rho_SOR_Q)
    np.savetxt('Jac.out', (rho_Jac_R, rho_Jac_Q))
    print timers_Q
    '''


if __name__ == "__main__":
    main()
