#include "katoptron.h"
#include "wMshData.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wDirichlet.h"
#include "wNeumann.h"
#include "wContact.h"
#include "IterativeSolver.h"
#include <mpi.h>

#include "wGmshImport.h"
#include "wNode.h"
#include "DiscreteProblem.h"

#include <iostream>
#include <cmath>

#include <Teuchos_ParameterList.hpp>
#include <Teuchos_RCP.hpp>

#include <stdlib.h>

#include <xmmintrin.h>
#include <cmath>

#include "Belos_Tpetra_MP_Vector.hpp"
#include "BelosLinearProblem.hpp"
#include "BelosPseudoBlockGmresSolMgr.hpp"

#include "Kokkos_Init.h"

const double pi = std::acos(-1);

template <typename scalar>
void test_function(int ensemble_size)
{

    typedef Tpetra::Vector<>::local_ordinal_type local_ordinal_type;
    typedef Tpetra::Vector<>::global_ordinal_type global_ordinal_type;

    //typedef Tpetra::Map<> map_type;
    //typedef Tpetra::Vector<>::scalar_type scalar_type;
    //typedef Tpetra::Vector<>::mag_type magnitude_type;
    typedef Tpetra::MultiVector<scalar, local_ordinal_type, global_ordinal_type> multivector_type;
    //typedef Tpetra::CrsMatrix<scalar,local_ordinal_type,global_ordinal_type> crs_matrix_type;
    //typedef Tpetra::CrsGraph<local_ordinal_type,global_ordinal_type> crs_graph_type;
    //typedef Tpetra::Vector<scalar,local_ordinal_type,global_ordinal_type> vector_type;

    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    int DOFperNode = 3;

    std::string mesh_name;
    mesh_name = "cube.msh";

    auto msh = std::make_shared<MshData>();
    auto gmshLoader = std::make_shared<GmshImport>(msh);
    if (worldsize > 1)
        gmshLoader->load(mesh_name, myrank);
    else
        gmshLoader->load(mesh_name, -1);

    //auto numNodes = msh->nodes.size();

    katoptron::Problem pbl(msh, MPI_COMM_WORLD);

    double E_value = 21000.;
    double nu_value = 0.29;
    double k = 0.;
    double d = 0.;

    std::vector<double> E, nu;
    for (auto i = 0; i < ensemble_size; ++i)
    {
        E.push_back(E_value);
        nu.push_back(nu_value);
    }

    katoptron::Medium md0(pbl, std::string("Body 1"), std::string("Mat 1"), E, nu, k, d);
    pbl.media[0] = &md0;

    std::vector<double> zero, dy;
    for (auto i = 0; i < ensemble_size; ++i)
    {
        zero.push_back((double)0.);
        dy.push_back((double)1.);
    }

    katoptron::Dirichlet ud0(pbl, std::string("Surf 1 1"), std::string("Clamped 1"), 1, zero, 1, zero, 1, zero, 0, zero);
    katoptron::Dirichlet ud1(pbl, std::string("Surf 1 2"), std::string("Clamped 1"), 0, zero, 1, dy, 0, zero, 0, zero);

    pbl.duBCs[0] = &ud0;
    pbl.duBCs[1] = &ud1;

    Teuchos::ParameterList solverList;

    scalar tol = (scalar)1E-06;
    solverList.set(std::string("Convergence Tolerance"), tol);
    solverList.set(std::string("Maximum Iterations"), 1000);
    solverList.set(std::string("Verbosity"), 33);
    solverList.set(std::string("Flexible Gmres"), false);
    solverList.set(std::string("Output Style"), 1);
    solverList.set(std::string("Output Frequency"), 1);
    //solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
    solverList.set(std::string("Output Style"), 1);
    //solverList.set(std::string("convert MueLu xml file"),           true);
    solverList.set(std::string("Use preconditioner"), false);
    solverList.set(std::string("Maximum active set iteration"), 1);
    solverList.set(std::string("type"), std::string("BlockGmres"));

    RCP<Teuchos::ParameterList> solverList2 = rcp(new Teuchos::ParameterList(solverList));

    RCP<Teuchos::ParameterList> randomParams = rcp(new Teuchos::ParameterList());
    //katoptron::IterativeSolver slv(pbl, 1, solverList2, DOFperNode, (ensemble_size > 1), ensemble_size);

    auto begin = std::chrono::high_resolution_clock::now();

    Kokkos::View<scalar *, Kokkos::LayoutLeft> a = Kokkos::View<scalar *, Kokkos::LayoutLeft>("KL Random Variables", 0);
    katoptron::DiscreteProblem<scalar> discreteProblem(pbl, DOFperNode, randomParams, a);

    discreteProblem.computeMatrices();
    discreteProblem.computeLoads();

    typedef multivector_type MV;
    typedef typename multivector_type::dot_type belos_scalar;
    typedef Tpetra::Operator<scalar> OP;
    typedef Belos::LinearProblem<belos_scalar, MV, OP> BLinProb;
    RCP<BLinProb> blinproblem = rcp(new BLinProb(discreteProblem.algebraic->matrices->K, discreteProblem.algebraic->vectors->x, discreteProblem.algebraic->vectors->b));
    blinproblem->setProblem();
    RCP<Teuchos::ParameterList> belosParams = rcp(new Teuchos::ParameterList(solverList));
    RCP<Belos::SolverManager<belos_scalar, MV, OP>> blinsolver =
        rcp(new Belos::PseudoBlockGmresSolMgr<belos_scalar, MV, OP>(blinproblem, belosParams));
    //Belos::ReturnType ret = blinsolver->solve();
    blinsolver->solve();

    auto end = std::chrono::high_resolution_clock::now();
    double total_time = std::chrono::duration_cast<std::chrono::nanoseconds>(end - begin).count();
    std::cout << ensemble_size << " " << total_time << " nanosec" << std::endl;
}

template <typename scalar>
void test_function2(int ensemble_size)
{
    int worldsize;
    MPI_Comm_size(MPI_COMM_WORLD, &worldsize);

    int myrank;
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

    int DOFperNode = 3;

    std::string mesh_name;
    mesh_name = "cube.msh";

    auto msh = std::make_shared<MshData>();
    auto gmshLoader = std::make_shared<GmshImport>(msh);
    if (worldsize > 1)
        gmshLoader->load(mesh_name, myrank);
    else
        gmshLoader->load(mesh_name, -1);

    //auto numNodes = msh->nodes.size();

    katoptron::Problem pbl(msh, MPI_COMM_WORLD);

    double E_value = 21000.;
    double nu_value = 0.29;
    double k = 0.;
    double d = 0.;

    std::vector<double> E, nu;
    for (auto i = 0; i < ensemble_size; ++i)
    {
        E.push_back(E_value);
        nu.push_back(nu_value);
    }

    katoptron::Medium md0(pbl, std::string("Body 1"), std::string("Mat 1"), E, nu, k, d);
    pbl.media[0] = &md0;

    std::vector<double> zero, dy;
    for (auto i = 0; i < ensemble_size; ++i)
    {
        zero.push_back((double)0.);
        dy.push_back((double)1.);
    }

    katoptron::Dirichlet ud0(pbl, std::string("Surf 1 1"), std::string("Clamped 1"), 1, zero, 1, zero, 1, zero, 0, zero);
    katoptron::Dirichlet ud1(pbl, std::string("Surf 1 2"), std::string("Clamped 1"), 0, zero, 1, dy, 0, zero, 0, zero);

    pbl.duBCs[0] = &ud0;
    pbl.duBCs[1] = &ud1;

    Teuchos::ParameterList solverList;

    solverList.set(std::string("Ensemble Convergence Tolerance"), 1E-06);
    solverList.set(std::string("Maximum Iterations"), 1000);
    solverList.set(std::string("Verbosity"), 33);
    solverList.set(std::string("Flexible Gmres"), false);
    solverList.set(std::string("Output Style"), 1);
    solverList.set(std::string("Output Frequency"), 1);
    //solverList.set(std::string("MueLu xml file name"),              std::string("nested_multigrid_2_lvls.xml"));
    solverList.set(std::string("Output Style"), 1);
    //solverList.set(std::string("convert MueLu xml file"),           true);
    solverList.set(std::string("Use preconditioner"), false);
    solverList.set(std::string("Maximum active set iteration"), 1);
    solverList.set(std::string("type"), std::string("BlockGmres"));
    solverList.set(std::string("Write vtk files"), false);

    RCP<Teuchos::ParameterList> solverList2 = rcp(new Teuchos::ParameterList(solverList));
    katoptron::IterativeSolver slv(pbl, solverList2, DOFperNode, ensemble_size);

    auto begin2 = std::chrono::high_resolution_clock::now();
    slv.start();
    auto end2 = std::chrono::high_resolution_clock::now();
    double total_time2 = std::chrono::duration_cast<std::chrono::nanoseconds>(end2 - begin2).count();
    std::cout << ensemble_size << " " << total_time2 << " nanosec" << std::endl;
}

int main(int argc, char **argv)
{
    size_t case_id, N, k;
    if (argc > 1)
        case_id = atoi(argv[1]);
    else
        case_id = 1;

    if (argc > 2)
        N = atoi(argv[2]);
    else
        N = 1;

    if (argc > 3)
        k = atoi(argv[3]);
    else
        k = 1;

    MPI_Init(NULL, NULL);
    Kokkos_Initialize(k);
    {
        const int ensemble_size = 1;
        typedef double scalar;
        for (size_t i = 0; i < N; ++i)
        {
            std::cout << " ==================== " << std::endl;
            std::cout << " =-- start test " << i << " --= " << std::endl;
            std::cout << " ==================== " << std::endl;
            if (case_id == 1)
                test_function<scalar>(ensemble_size);
            else
                test_function2<scalar>(ensemble_size);
        }
    }
    Kokkos_Finalize();
    MPI_Finalize();
}
