/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @brief Return wether the element values have been initialized or not
 */
inline bool Element::hasValues() const
{
    return hasVals;
}

/**
 * @brief Return wether the element gradients have been initialized or not
 */
inline bool Element::hasGradients() const
{
    return hasGrads;
}

/**
 * @brief Return the determinant of the Jacobian at Gauss point k
 */
inline double Element::getDetJ(size_t k) const
{
    if (hasVals)
        return detJs[k];
    else
        throw std::runtime_error("Element::getDetJ is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the the Jacobian matrix at Gauss point k
 */
inline Eigen::MatrixXd const &Element::getJ(size_t k) const
{
    if (hasVals)
    {
        if (isVol)
            return Js[k];
        else
            throw std::runtime_error("Element::getJ Jacobian matrix is not available for surface element!\n");
    }
    else
        throw std::runtime_error("Element::getJ is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the inverse of the Jacobian matrix at Gauss point k
 */
inline Eigen::MatrixXd const &Element::getJinv(size_t k) const
{
    if (hasVals)
    {
        if (isVol)
            return iJs[k];
        else
            throw std::runtime_error("Element::getJinv inverse Jacobian matrix is not available for surface element!\n");
    }
    else
        throw std::runtime_error("Element::getJinv is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the volume of the element
 */
inline double Element::getVol() const
{
    if (hasVals)
        return vol;
    else
        throw std::runtime_error("Mem::getVol is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the center of gravity of the element
 */
inline Eigen::Vector3d const &Element::getCg() const
{
    if (hasVals)
        return cg;
    else
        throw std::runtime_error("Element::getCg is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the unit normal vector of the element
 */
inline Eigen::Vector3d const &Element::getNormal() const
{
    if (hasVals)
    {
        if (!isVol)
            return normal;
        else
            throw std::runtime_error("Element::getNormal normal vector is not available for volume element!\n");
    }
    else
        throw std::runtime_error("Element::getNormal is invalid because element memory is not up-to-date!\n");
}

/**
 * @brief Return the gradient of determinant of the Jacobian with respect to node positions at Gauss point k
 */
inline std::vector<double> const &Element::getGradDetJ(size_t k) const
{
    if (hasGrads)
        return gradDetJs[k];
    else
        throw std::runtime_error("Element::getGradDetJ is invalid because element memory (gradient) is not up-to-date!\n");
}

/**
 * @brief Return the gradient of the Jacobian matrix with respect to node positions at Gauss point k
 */
inline std::vector<Eigen::MatrixXd> const &Element::getGradJ(size_t k) const
{
    if (hasGrads)
    {
        if (isVol)
            return gradJs[k];
        else
            throw std::runtime_error("Element::getGradJ gradient of Jacobian matrix is not available for surface element!\n");
    }
    else
        throw std::runtime_error("Element::getGradJ is invalid because element memory (gradient) is not up-to-date!\n");
}

/**
 * @brief Return the gradient of the volume with respect to node positions
 */
inline std::vector<double> const &Element::getGradVol() const
{
    if (hasGrads)
        return gradVol;
    else
        throw std::runtime_error("Element::getGradVol is invalid because element memory (gradient) is not up-to-date!\n");
}

/**
 * @brief Return the gradient of the unit normal vector with respect to node positions
 */
inline std::vector<Eigen::Vector3d> const &Element::getGradNormal() const
{
    if (hasGrads)
    {
        if (!isVol)
            return gradNormal;
        else
            throw std::runtime_error("Element::getGradNormal normal vector gradient is not available for volume element!\n");
    }
    else
        throw std::runtime_error("Element::getGradNormal is invalid because element memory (gradient) is not up-to-date!\n");
}
