#include "wExample9.h"
/*
#@HEADER
# ************************************************************************
#
#                          Moertel FE Package
#                 Copyright (2006) Sandia Corporation
#
# Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
# license for use of this work by or on behalf of the U.S. Government.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the Corporation nor the names of the
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Contact Glen Hansen (gahanse@sandia.gov)
#
# ************************************************************************
#@HEADER
*/
/*!
 * \file TwoSquares.cpp
 *
 * \brief Simple serial example showing Moertel usage and solver interfaces
 *
 * \date Last update do Doxygen: 20-March-06
 *
 */
#ifdef HAVE_MPI
#include "mpi.h"
#include "Epetra_MpiComm.h"
#else
#include "Epetra_SerialComm.h"
#endif

// MOERTEL headers
#include "mrtr_manager.H"
#include "mrtr_segment_linear1D.H"

// Galeri headers
#include "Galeri_Utils.h"
#include "Galeri_FiniteElements.h"
#ifdef MOERTEL_HAVE_EXODUS
#include "ExodusInterface.h"
#endif
using namespace Galeri;
using namespace Galeri::FiniteElements;

// ==========================================================
// This file solves the scalar problem
//
//   - \mu \nabla u + \sigma u = f    on \Omega
//                           u = g    on \partial \Omega
//
// where \Omega is a 2D rectangle, divided into triangles.
// The input grid should be generated using similar
// conventions of file galeri/data/TwoSquares.m:
// - the bc ID of 10 and 20 are for the mortar interface
// - the bc ID of 0 is for the external domain.
// ==========================================================

double Diffusion9(const double &x, const double &y, const double &z)
{
    return (1.0);
}

double Source9(const double &x, const double &y, const double &z)
{
    return (0.0);
}

double Force9(const double &x, const double &y, const double &z)
{
    return (0.0);
}

// Specifies the boundary condition.
int BoundaryType9(const int &Patch)
{
    if (Patch == 11 || Patch == 12)
        return (GALERI_DIRICHLET);
    else
        return (GALERI_DO_NOTHING);
}

// Specifies the boundary condition.
double BoundaryValue9(const double &x, const double &y,
                      const double &z, const int &Patch)
{
    if (Patch == 11 || Patch == 12)
        return (-y / 4.0 + 0.75);
    else
        return (1.0);
}

// =========== //
// main driver //
// =========== //

using namespace tlnos;

Example9::Example9()
{
}

int Example9::execute()
{
#ifdef HAVE_MPI
    /*
   static int  argc = 1;
   char **argv = new char*[2];// mem leak volontaire
   argv[0] = new char[8]; strcpy(argv[0], "myprog"); // mem leak volontaire
   argv[1] = new char[1]; strcpy(argv[1], ""); // mem leak volontaire
  MPI_Init(&argc,&argv);
*/

    Epetra_MpiComm Comm(MPI_COMM_WORLD);
#else
    Epetra_SerialComm Comm;
#endif

    int status = 0; // return status

    try
    {

        // this example is in serial only
        if (Comm.NumProc() > 1)
            exit(0);

        // read grid from file, see also TwoSquares.m used to generate the grids
        /*
    %  +------+
    %  |  S2  |
    %  +------+ <- edge has tag 20 and 9 elements
    %  +------+ <- edge has tag 10 and 10 elements
    %  |  S1  |
    %  +------+
    %
    % where S1 = (-1,1) x (-1,1) and S2 = (-1, 1) x (1, 3).
    */

        FileGrid Grid(Comm, "TwoSquaresDirichlet.grid");

        // create a list of all nodes that are linked to a face
        // with tag 10 and tag 20
        map<int, int> nodes10;
        map<int, int> nodes20;
        for (int i = 0; i < Grid.NumMyBoundaryFaces(); ++i)
        {
            int tag;
            int nodeids[2];
            Grid.FaceVertices(i, tag, nodeids);
            if (tag == 10)
            {
                nodes10[nodeids[0]] = nodeids[0];
                nodes10[nodeids[1]] = nodeids[1];
            }
            else if (tag == 20)
            {
                nodes20[nodeids[0]] = nodeids[0];
                nodes20[nodeids[1]] = nodeids[1];
            }
            else
                continue;
        }

        // ------------------------------------------------------------- //
        // create an empty MOERTEL::Interface, in this example just one
        // ------------------------------------------------------------- //
        int printlevel = 0; // ( moertel takes values 0 - 10 )
        //int printlevel = 9; // ( moertel takes values 0 - 10 )
        MOERTEL::Interface interface(0, true, Comm, printlevel);

        // ------------------------------------------------------------- //
        // Add nodes on both sides of interface to interface
        // loop all nodes in the maps nodes10 and nodes20 and add them
        // to the interface with unique ids
        // tag 10 will become interface side 0
        // tag 20 will become interface side 1
        // ------------------------------------------------------------- //
        map<int, int>::iterator curr;
        // do tag==10 or interface side 0
        for (curr = nodes10.begin(); curr != nodes10.end(); ++curr)
        {
            // get unique node id (here it's equal to the degree of freedom on that node)
            int nodeid = curr->second;
            // get node coordinates
            double coord[3];
            Grid.VertexCoord(nodeid, coord);
            // get dirichlet boundary conditions
            double bou = BoundaryValue9(coord[0], coord[1], coord[2], 10);
            bool dboundary = false;
            if (bou == 0.0)
            {
                dboundary = true;
                std::cout << "hi" << std::endl;
            }
            // create a moertel node
            MOERTEL::Node node(nodeid, coord, 1, &nodeid, dboundary, printlevel);
            // add node to the interface on side 0
            interface.AddNode(node, 0);
        }
        nodes10.clear();

        // do tag==20 or interface side 1
        for (curr = nodes20.begin(); curr != nodes20.end(); ++curr)
        {
            // get unique node id (here it's equal to the degree of freedom on that node)
            int nodeid = curr->second;
            // get node coordinates
            double coord[3];
            Grid.VertexCoord(nodeid, coord);
            // get dirichlet boundary conditions
            double bou = BoundaryValue9(coord[0], coord[1], coord[2], 20);
            bool dboundary = false;
            if (bou == 0.0)
            {
                dboundary = true;
                std::cout << "hi" << std::endl;
            }
            // create a moertel node
            MOERTEL::Node node(nodeid, coord, 1, &nodeid, dboundary, printlevel);
            // add node to the interface on side 1
            interface.AddNode(node, 1);
        }
        nodes20.clear();

        // ------------------------------------------------------------- //
        // add segments on both sides of the interface to the interface
        // ------------------------------------------------------------- //
        for (int i = 0; i < Grid.NumMyBoundaryFaces(); ++i)
        {
            int tag;
            int nodeids[2];
            Grid.FaceVertices(i, tag, nodeids);
            if (tag != 10 && tag != 20)
                continue;
            // create a segment (galeri calls it a face)
            MOERTEL::Segment_Linear1D segment(i, 2, nodeids, printlevel);

            // add it to the interface on side 0
            if (tag == 10)
                interface.AddSegment(segment, 0);
            // add it to the interface on side 1
            else if (tag == 20)
                interface.AddSegment(segment, 1);
        }

        // ------------------------------------------------------------- //
        // choose the mortar side of the interface (0 or 1)
        // here: let the package choose it (-2)
        // ------------------------------------------------------------- //
        interface.SetMortarSide(-2);

        // ------------------------------------------------------------- //
        // As we do not know the mortar side yet (we decided to le the
        // package choose it), we can not set a dual trace function (mortar space)
        // as we don't know the side to set it to
        // so we just give orders for the function type
        // ------------------------------------------------------------- //
        interface.SetFunctionTypes(MOERTEL::Function::func_Linear1D, // primal trace space
                                   //MOERTEL::Function::func_DualLinear1D);  // dual mortar space (recommended)
                                   MOERTEL::Function::func_Linear1D); // mortar space (not recommended)

        // ------------------------------------------------------------- //
        // complete the interface
        // ------------------------------------------------------------- //
        if (!interface.Complete())
        {
            cout << "Interface completion returned false\n";
            exit(EXIT_FAILURE);
        }

        // ------------------------------------------------------------- //
        // create an empty MOERTEL::Manager for 2D problems
        // It organizes everything from integration to solution
        // ------------------------------------------------------------- //
        MOERTEL::Manager manager(Comm, printlevel);
        manager.SetDimension(MOERTEL::Manager::manager_2D);

        // ------------------------------------------------------------- //
        // Add the interface to the manager
        // ------------------------------------------------------------- //
        manager.AddInterface(interface);

        // ------------------------------------------------------------- //
        // for mortar integration, the mortar manager needs to know about
        // the rowmap of the original (uncoupled) problem because it will
        // create coupling matrices D and M matching that rowmap
        // ------------------------------------------------------------- //
        manager.SetProblemMap(&Grid.RowMap());

        // ============================================================= //
        // choose integration parameters
        // ============================================================= //
        Teuchos::ParameterList &moertelparams = manager.Default_Parameters();
        // this does not affect this 2D case
        moertelparams.set("exact values at gauss points", true);
        // 1D interface possible values are 1,2,3,4,5,6,7,8,10 (2 recommended with linear shape functions)
        moertelparams.set("number gaussian points 1D", 2);
        // 2D interface possible values are 3,6,12,13,16,19,27
        moertelparams.set("number gaussian points 2D", 27);

        // ============================================================= //
        // Here we are done with the construction phase of the interface
        // so we can integrate the mortar integrals
        // (Note we have not yet evaluated the PDE at all!)
        // ============================================================= //
        manager.Mortar_Integrate();

        // print interface information
        // (Manager, Interface, Segment, Node implement the << operator)
        if (printlevel)
            cout << manager;

        // ======================================================== //
        // Prepares the linear system. This requires the definition //
        // of a quadrature formula compatible with the grid, a      //
        // variational formulation, and a problem object which take //
        // care of filling matrix and right-hand side.              //
        // NOTE:
        // we are doing this AFTER we did all the mortar stuff to
        // show that the mortar integration is actually PDE-independent
        // ======================================================== //
        Epetra_CrsMatrix A(Copy, Grid.RowMap(), 0);
        Epetra_Vector LHS(Grid.RowMap(), true);
        Epetra_Vector RHS(Grid.RowMap());

        int NumQuadratureNodes = 3;

        GalerkinVariational<TriangleQuadrature>
            Laplace2D(NumQuadratureNodes, Diffusion9, Source9, Force9,
                      BoundaryValue9, BoundaryType9);

        LinearProblem FiniteElementProblem(Grid, Laplace2D, A, LHS, RHS);
        FiniteElementProblem.Compute();

        // ============================================================= //
        // this is Galeri's dense solve method if you'd like to see how
        // the uncoupled solution looks like
        // ============================================================= //
        Solve(&A, &LHS, &RHS);
#ifdef MOERTEL_HAVE_EXODUS
        ExodusInterface exodus(Comm);
        exodus.Write(Grid, "uncoupled_output", LHS);
#endif
        // ============================================================= //
        // Pass input matrix to Moertel,
        // Moertel does NOT take ownership of A!
        // ============================================================= //
        manager.SetInputMatrix(&A, false);

        // ============================================================= //
        // Solve
        // ============================================================= //

        Epetra_CrsMatrix *B = manager.MakeSaddleProblem();

        Epetra_CrsMatrix B1(*B);
        Epetra_CrsMatrix B2(Copy, (*B).RowMap(), 0);
        Epetra_Vector LHS2(B2.RowMap(), true);
        Epetra_Vector LHS3(A.RowMap(), true);
        Epetra_Vector RHS2(B2.RowMap());

        for (size_t i = 0; i < A.NumGlobalRows(); ++i)
            RHS2[i] = RHS[i];

        int B1NumMyRow = B1.NumMyRows();
        int B1MaxNumIndices = B1.MaxNumEntries();
        double *values = new double[B1MaxNumIndices];
        int B1NumIndices;
        int *Indices_int = 0;
        Indices_int = new int[B1MaxNumIndices];

        for (size_t i = 0; i < B1NumMyRow; ++i)
        {
            B1.ExtractGlobalRowCopy(i, B1MaxNumIndices, B1NumIndices, values, Indices_int);
            B2.InsertGlobalValues(i, B1NumIndices, values, Indices_int);
        }

        double U;
        int i_treshold = (B2.NumGlobalRows() + A.NumGlobalRows()) / 2;
        for (int i = A.NumGlobalRows(); i < B2.NumGlobalRows(); ++i)
        {
            if (i < i_treshold)
                U = 0.;
            else
                U = 10000.;
            B2.InsertGlobalValues(i, 1, &U, &i);
        }

        B2.FillComplete();

        Solve(&B2, &LHS2, &RHS2);

        for (size_t i = 0; i < A.NumGlobalRows(); ++i)
            LHS3[i] = LHS2[i];
#ifdef MOERTEL_HAVE_EXODUS
        exodus.Write(Grid, "coupled_output", LHS3);
#endif
        std::cout << "Lambda" << std::endl
                  << "-----" << std::endl;

        for (size_t i = A.NumGlobalRows(); i < B2.NumGlobalRows(); ++i)
            std::cout << LHS2[i] << std::endl;

        std::cout << *(manager.M()) << std::endl
                  << "-----" << std::endl;
        std::cout << *(manager.D()) << std::endl
                  << "-----" << std::endl;
    }
    catch (int e)
    {
        cerr << "Caught exception, value = " << e << endl;
        status = 1;
    }
    catch (Galeri::Exception &rhs)
    {

        cerr << "Caught Galeri exception: ";
        rhs.Print();
        status = 1;
    }
    catch (...)
    {
        cerr << "Caught generic exception" << endl;
        status = 1;
    }

#ifdef HAVE_MPI
    MPI_Finalize();
#endif

    if (status == 0)
        std::cout << "\nTest passed!" << endl;
    else
        std::cout << "\nTest Failed!" << endl;

    // Final return value (0 = successfull, non-zero = failure)

    return status;
}
